#!/usr/bin/perl
#
# fixTCPhang.pl
#
# script for detecting hung TCP connections and, optionally, clearing them
#
# August 2004, Chris Elliott
#
# This code written by me for cisco customers and freely made available 
#  with no warranty or support.
# 
# Please NOTE:  None of this code available here comes with any
#  warranty or support.
#
# History: 
#  Date               Rev  Who                               What
#  August 26, 2004    1.0  Chris Elliott chelliot@cisco.com  Initial Coding
#  September 1, 2004  1.1  Chris Elliott chelliot@cisco.com  Added indication that TCP connection had been cleared and changed MIB path option
#  September 2, 2004  1.2  Chris Elliott chelliot@cisco.com  Added support for SNMP versions

use Getopt::Std;
      
getopts('c:C:dfhm:p:t:v:V');

$Version = "1.2";
if (index($0,"/") > -1) {
    $ProgName = substr($0,rindex($0,"/")+1);
} else {
    $ProgName = $0;
}

if ($ARGV[0] eq "" || $opt_h) {
  die "$ProgName [-dfhV -c <read_community> -C <write_community> -m <mib_directory> -p <command_path> -t <timeout> -v <snmp_version>] <device>\n",
      "  Version $Version\n\n",
      "  Detect hung TCP connections on <device>, optionally clearing them.\n",
      "  Options:  -c  Specify read community string. Defaults to public.\n",
      "            -C  Specify the readwrite community string. No default.\n",
      "                 Must be supplied for the script to clear hung connections.\n",
      "            -d  Turn on debug mode.\n",
      "            -f  Fix or clear any hung TCP connections found.\n",
      "            -h  Print this message.\n",
      "            -m  Specify the directory to find CISCO-SMI.my and CISCO-TCP-MIB.my.\n",
      "                 Defaults to /tmp/mibs.\n",
      "            -p  Where to find the net-snmp utilities.\n",
      "                 Optional if the utilities are in the path.\n",
      "            -t  SNMP Timeout value. Defaults to 5 sec.\n",
      "            -v  Specify SNMP version to use: One of 1, 2c, or 3.\n",
      "                 If 3 is specified then this option must include all of the\n",
      "                 authentication information for SNMPv3. For example:\n",
      "                 \"3 -a MD5 -u chelliot -A chelliot -l authNoPriv\"\n",
      "                 Note: NET-SNMP seems to have a bug with SHA authentication on Windows.\n",
      "                 See the NET-SNMP documentation for more information.\n",
      "                 Defaults to SNMP version 1.\n",
      "            -V  Print version number.\n";
}

$device = $ARGV[0];

print "$ProgName Version $Version\n\n" if $opt_V;
print STDERR "$ProgName: Version $Version\n\n" if $opt_d;

# Use the read-write community string for read if one is supplied unless a read 
#  community string is supplied.
# Read community string defaults to public otherwise.
$opt_c = $opt_C if (defined($opt_C) && !defined($opt_c));
$opt_c = "public" if (!defined($opt_c));
$opt_C = $opt_c if (!defined($opt_C));

# We need MIBs...
$opt_m = "/tmp/mibs" if (!$opt_m);
chop($opt_m) if (substr($opt_m,length($opt_m),1) eq "/" || substr($opt_m,length($opt_m),1) eq "\\");
if ($ENV{'OS'} eq "Windows_NT") {
    $opt_m = "-m $opt_m/CISCO-SMI.my -m $opt_m/CISCO-TCP-MIB.my";
} else {
    $opt_m = "-m $opt_m/CISCO-SMI.my:$opt_m/CISCO-TCP-MIB.my";
}

if ($opt_p) {
    $opt_p = "$opt_p/"  if (substr($opt_p,length($opt_p),1) ne "/");
}
$snmpgetnext = $opt_p."snmpgetnext";
$snmpget = $opt_p."snmpget";
$snmpset = $opt_p."snmpset";

$opt_t = 5 if (!$opt_t);

$opt_v = "1" if (!$opt_v);

# Arguments to pass to net-snmp--use SNMPv1 for compatibility with old devices...
$SNMPargs = "-t $opt_t $opt_m -v $opt_v -O S";

# Determine if ciscoTcpConnRetransPkts and ciscoTcpConnRto are supported in the device.
$doSix = 0;
$result = `$snmpgetnext $SNMPargs -c $opt_c $device ciscoTcpConnRto` || die "Error: Can't get ciscoTcpConnRto via SNMP, $? $!\n";
print STDERR "$ProgName: determine if all objects are available: $result\n" if ($opt_d);
($mib,$sixTest,$rest) = split(/[ \t\n:=\.]+/,$result,3);
$doSix = 1 if ($sixTest eq "ciscoTcpConnRto");
print STDERR "$ProgName: Only use byte and packet counts.\n" if (!$doSix && $opt_d);
print STDERR "$ProgName: Use all six objects.\n" if ($doSix && $opt_d);

# Setup the list of objects to poll
$oids = "ciscoTcpConnInBytes ciscoTcpConnOutBytes ciscoTcpConnOutBytes ciscoTcpConnOutPkts";
$oids = "$oids ciscoTcpConnRetransPkts ciscoTcpConnRto" if ($doSix);

$stop = 0;

do {
# Get a row of objects

    $result = `$snmpgetnext $SNMPargs -c $opt_c $device $oids` || die "Error: Can't get $oids via SNMP, $? $!\n";
    ($inBytes,$inPkts,$outBytes,$outPkts,$retransPkts,$rto) = split(/\n/,$result);

# Parse the Out Bytes
    ($mib,$OBoid,$type,$OBval,$units) = split(/[ \t\n:=]+/,$outBytes,5);
    ($OB,$OBbaseoid) = split(/\./,$OBoid,2);

# If the following test isn't true then we've walked all of the entries in the table
    if ($OB ne "ciscoTcpConnOutBytes") {
	$stop = 1;
    } else {
	@OID = split(/\./,$OBoid);
        print STDERR "$ProgName: TCP Connection: Local IP: @OID[1].@OID[2].@OID[3].@OID[4] port @OID[5] Remote IP: @OID[6].@OID[7].@OID[8].@OID[9] port @OID[10]\n" if ($opt_d);

# Parse the In Bytes
        ($mib,$IBoid,$type,$IBval,$units) = split(/[ \t\n:=]+/,$inBytes,5);
        ($IB,$IBbaseoid) = split(/\./,$IBoid,2);
        print STDERR "$ProgName: In Bytes = $IBval.\n" if ($opt_d);

# Parse the In Packets
	($mib,$IPoid,$type,$IPval,$units) = split(/[ \t\n:=]+/,$inPkts,5);
	($IP,$IPbaseoid) = split(/\./,$IPoid,2);
	print STDERR "$ProgName: In Packets = $IPval.\n" if ($opt_d);
	print STDERR "$ProgName: Out Bytes = $OBval.\n" if ($opt_d);

# Parse the Out Packets
       ($mib,$OPoid,$type,$OPval,$units) = split(/[ \t\n:=]+/,$outPkts,5);
       ($OP,$OPbaseoid) = split(/\./,$OPoid,2);
       print STDERR "$ProgName: Out Packets = $OPval.\n" if ($opt_d);

       if ($doSix) {

# Parse the Retransmitted Packets
	   ($mib,$RPoid,$type,$RPval,$units) = split(/[ \t\n:=]+/,$retransPkts,5);
	   ($RP,$RPbaseoid) = split(/\./,$RPoid,2);

	   print STDERR "$ProgName: Retransmitted Packets = $RPval.\n" if ($opt_d);

# Parse the Retransmission Timeout value
	   ($mib,$RTOoid,$type,$RTOval,$units) = split(/[ \t\n:=]+/,$rto,5);
	   ($RTO,$RTObaseoid) = split(/\./,$RTOoid,2);
	   print STDERR "$ProgName: Retransmission Timeout = $RTOval.\n" if ($opt_d);
       }

       if ($OBbaseoid ne $IBbaseoid || $OBbaseoid ne $IPbaseoid || $OBbaseoid ne $OPbaseoid || ($doSix && ($OBbaseoid ne $RPbaseoid || $OBbaseoid ne $RTObaseoid))) {
	   die "Error: Objects out of sync. Please retry the command.\n";
       }

# Determine OIDs for next getnext operation
       $oids = "$IBoid $IPoid $OBoid $OPoid";
       $oids = "$oids $RPoid $RTOoid" if ($doSix);

# Determine if we might have a bad TCP connection
       if (($doSix || ($IBval > 0 && $IPval == 0)) && $OBval == 0 && $OPval == 0 && (!$doSix || ($RPval > 0 && $RTOval > 20000))) {

           if (!$doSix) {
# Be cautious if we don't have all six objects to compare--sleep for 5 seconds and verify
	       sleep 5;

	       $result = `$snmpget $SNMPargs -c $opt_c $device $IBoid $IPoid $OBoid $OPoid` || die "Error: Can't get $IBoid $IPoid $OBoid $OPoid via SNMP, $? $!\n";
               ($inBytes,$inPkts,$outBytes,$outPkts) = split(/\n/,$result);
	       ($mib,$IBoid,$type,$IBval,$units) = split(/[ \t\n:=]+/,$inBytes,5);
	       ($mib,$IPoid,$type,$IPval,$units) = split(/[ \t\n:=]+/,$inPkts,5);
	       ($mib,$OBoid,$type,$OBval,$units) = split(/[ \t\n:=]+/,$outBytes,5);
	       ($mib,$OPoid,$type,$OPval,$units) = split(/[ \t\n:=]+/,$outPkts,5);
	       print STDERR "$ProgName: InBytes=$IBval,InPkts=$IPval,OutBytes=$OBval,OutPkts=$OPval,second time around.\n" if ($opt_d);
	   }
	   if (($doSix || ($IBval > 0 && $IPval == 0)) && $OBval == 0 && OPval ==0) {
	       print "Found bad TCP connection: Local IP: @OID[1].@OID[2].@OID[3].@OID[4] port @OID[5] Remote IP: @OID[6].@OID[7].@OID[8].@OID[9] port @OID[10]";
	       if ($opt_f) {
		   $result=`$snmpset $SNMPargs -c $opt_C $device tcpConnState.@OID[1].@OID[2].@OID[3].@OID[4].@OID[5].@OID[6].@OID[7].@OID[8].@OID[9].@OID[10] i 12`  || die "Error: Can't set tcpConnState.@OID[1].@OID[2].@OID[3].@OID[4].@OID[5].@OID[6].@OID[7].@OID[8].@OID[9].@OID[10] via SNMP, $? $!\n";
                   print ": CLEARED";
	       }
               print "\n";
	       print STDERR "$ProgName: $result\n" if ($opt_d);
	   }
       }
   }
} until ($stop == 1); 

print STDERR "$ProgName: Done.\n" if ($opt_d);

exit;

